 /*****************************************************************************
 *   Copyright (C) 2004 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#ifndef ITUNESDB_H
#define ITUNESDB_H

#include <qdatetime.h>
#include <qfile.h>
#include <qstringlist.h>

#include "playlist.h"
#include "playlistitem.h"
#include "track.h"
#include "ipodsysinfo.h"
#include "utils.h"

#define RECENTLY_PLAYED_LIST_NAME "KPOD:Recently Played"

/**
@author Michael Schulze
*/

class ITunesDBTrack : public itunesdb::Track, itunesdb::utils::NonCopyAble {

    friend class ITunesDB;

    class ITunesDB& m_itunesdb;

    ITunesDBTrack( Q_UINT32 trackid, ITunesDB& itunesdb );
    ITunesDBTrack( const ITunesDBTrack& src );

public:
    virtual ~ITunesDBTrack();

    void setDateAdded( Q_UINT32 date );

    /**
     * Returns the file on the iPod this Track instance refers to.
     * @deprecated this really is IPod business
     */
    QString getFilePath() const;

protected:
    virtual void setDirty( bool dirtyFlag = true );
};


/**
 * This class represents all the rules and limits making up a smart playlist.
 * see @c itunesdb::SmartPlaylistRuleSet for further information.
 */
class ITunesDBSPLRuleSet
    : public itunesdb::SmartPlaylistRuleSet,
      public itunesdb::TrackPredicate,
      itunesdb::utils::NonCopyAble
{

    const class ITunesDB& mItunesdb;
    class ITunesDBPlaylist& mPlaylist;

    itunesdb::TrackComparatorPtr getComparatorFor( itunesdb::SPLLimitSort limitSort ) const;

    void applyLimits( itunesdb::TrackPtrList& trackList ) const;

    uint mLastSeenItunesDBVersion;
    uint mLastSeenRuleSetVersion;
    uint mRuleSetVersion;

protected:

    virtual itunesdb::SmartPlaylistRule * createStringRule(
        itunesdb::SPLStringField field,
        itunesdb::SPLStringAction action,
        const QString& testString );

    virtual itunesdb::SmartPlaylistRule * createUIntRule(
        itunesdb::SPLUIntField field, itunesdb::SPLUIntAction action,
        Q_UINT64 fromValue, Q_UINT64 toValue,
        Q_INT64  fromDate,  Q_INT64  toDate,
        Q_UINT64 fromUnits, Q_UINT64 toUnits );

    virtual bool operator() ( const itunesdb::Track * track ) const;

public:

    ITunesDBSPLRuleSet( const itunesdb::SmartPlaylistRuleSet& src,
                        ITunesDBPlaylist& playlist, const ITunesDB& itunesdb );

    ITunesDBSPLRuleSet( ITunesDBPlaylist& playlist, const ITunesDB& itunesdb );

    virtual ~ITunesDBSPLRuleSet();

    /**
     * adds a string based rule to the rule set and returns true if successful.
     */
    virtual bool addStringRule( itunesdb::SPLStringField field,
                 itunesdb::SPLStringAction action, const QString& testString, int position = -1 );

    /**
     * adds a uint based rule to the rule set and returns true if successful.
     */
    virtual bool addUIntRule( itunesdb::SPLUIntField field,
                              itunesdb::SPLUIntAction action,
                              Q_UINT64 fromValue, Q_UINT64 toValue,
                               Q_INT64 fromDate,   Q_INT64 toDate,
                              Q_UINT64 fromUnits, Q_UINT64 toUnits,
                              int position = -1 );

    /**
     * Refreshes the playlist associated with this set of rules. If the database didn't
     * change since the last refresh and @c force is false the playlist won't be updated.
     * @param force forces the rule set to update the playlist even if the database didn't change.
     */
    void refreshPlaylist( bool force = false );

};


/**
 * Special playlist implementation to use with the ITunesDB class. This adds important
 * functionalities like setting the sort order and updating the contents of smart playlists.<br>
 * To create a new ITunesDBPlaylist instance use the
 * @c ITunesDB::createNewPlaylist( const QString& ) factory method and use the
 * @c ITunesDB::addPlaylist( ITunesDBPlaylist* ) to add it to the database.<br>
 * Here is an example on how to create a "recently played" Smart Playlist:
 * <pre><code>
 *  // create a new playlist
 *  ITunesDBPlaylist * recentlyPlayed = itunesdb.createNewPlaylist( "recently played" );
 *  recentlyPlayed->setSortOrder( Playlist::SORTORDER_TIME_PLAYED );
 *
 *  // make it a smart playlist and get the ruleset
 *  SmartPlaylistRuleSet& ruleSet = recentlyPlayed->enableSmartPlaylist();
 *
 *  // add the rule "last played in the last 2 weeks"
 *  ruleSet.addInTheLastRule( FIELD_LASTPLAYED, false, -2, IN_THE_LAST_WEEKS );
 *
 *  // limit to the first 30 songs, sorted by last played date
 *  ruleSet.setLimits( LIMIT_SORT_LASTPLAYED, LIMIT_TYPE_SONGS, 30 );
 *
 *  // match ANY rule (doesn't really matter since we only have one rule)
 *  ruleSet.setMatchAnyFlag( true );
 *
 *  // enable live update (updates the playlist on the go)
 *  ruleSet.setLiveUpdateFlag( true );
 *
 *  // finally add the playlist to the database
 *  itunesdb.addPlaylist( recentlyPlayed );
 * </code></pre>
 */
class ITunesDBPlaylist : public itunesdb::Playlist {

    friend class ITunesDB;

    ITunesDB& m_itunesdb;
    ITunesDBPlaylist( const ITunesDBPlaylist& );
    ITunesDBPlaylist( ITunesDB& itunesdb );
    ITunesDBPlaylist( ITunesDB& itunesdb, const itunesdb::Playlist& playlist );
    itunesdb::TrackComparatorPtr getTrackComparatorFor( Q_UINT32 sortOrder );

    /**
     * @see @c itunesdb::ListItem
     */
    void doPostParsingChecks();

public:

    /**
     * An entry in this playlist. It directly refers to an existing track in the
     * database.
     */
    class Entry : public itunesdb::PlaylistItem {
        friend class ITunesDBPlaylist;
        unsigned int m_mhitTrackPos;
        Entry( Q_UINT32 id, uint trackpos );
    public:
        virtual ~Entry();

        unsigned int getMHITTrackPos() const {
            return m_mhitTrackPos;
        }
    };

    virtual ~ITunesDBPlaylist();

    /**
     * Creates a new instance of the Entry class.
     */
    virtual itunesdb::PlaylistItem * createNewItem( Q_UINT32 itemid );

    /**
     * Updates the content of this smart playlist against the database
     * Nothing happens if this playlist is not a smart playlist.
     * @see itunesdb::Playlist::enableSmartPlaylist()
     * @see itunesdb::Playlist::getSmartPlaylistRules()
     */
    void updateSmartPlaylist( bool force = false );

    /**
     * Sets the sort order field.
     * @see itunesdb::Playlist::ItdbPlaylistSortOrder
     */
    void setSortOrder( Sortorder sortOrder );

    /**
     * Returns the track the playlist entry at the given position is referring to
     */
    ITunesDBTrack * getTrackAt( uint position ) const;

    /**
     * Returns the track for a given playlist item
     * @param plItem the PlaylistItem to get the referred track 
     * @return the track for the given playlist item
     */
    ITunesDBTrack * getTrack( const itunesdb::PlaylistItem& plItem ) const;

protected:

    /**
     * reads mhods with type above type 15 (mhods not containing strings)
     */
    virtual void readNonStringMHOD( QDataStream& stream, Q_UINT32 type, Q_UINT32 blocklen );

    virtual void setDirty( bool dirtyFlag = true );

    virtual itunesdb::SmartPlaylistRuleSet * createNewSplRuleSet();

};


/**
 * Class representing the contents of your iPods ITunesDB file
 */
class ITunesDB : itunesdb::utils::NonCopyAble {
    class ItunesDBReaderAdapter;    // Adapter class for reading the itunesdb file
    class ItunesDBWriterAdapter;    // Adapter class for writing the itunesdb file

    // friends
    friend class ITunesDBPlaylist;
    friend class ITunesDBSPLRuleSet;
    friend class ITunesDBTrack;

    typedef std::map<Q_UINT32,uint> TrackIDMap;

    bool mDoDirtyFlagHandling;
    uint mInternalVersion;

    // TODO implement shared PImpl thus making it copyable

public:

    typedef itunesdb::utils::SortablePtrVector<ITunesDBTrack> TrackList;

    /**
     * @brief An Iterator over a sequence of ITunesDBTrack instances
     *
     * An Iterator over all ITunesDBTrack instances matching the given predicate, delivering pointers
     * to ITunesDBTrack instances by its next() method
     */
    template <typename TrackPredicate_T = itunesdb::TrackPredicates::AllTracks>
    struct FilteredTrackConstIterator
        : public TrackList::FilteredConstIterator<TrackPredicate_T> {

        /**
         * @brief Creates a new FilteredTrackConstIterator over all ITunesDBTrack instances matching the given predicate.
         * @param container the main TrackList to iterate over
         * @param predicate the predicate to match the tracks against
         */
        FilteredTrackConstIterator( const TrackList& tracklst, const TrackPredicate_T& pred =
            TrackPredicate_T() )
                : TrackList::FilteredConstIterator<TrackPredicate_T>( tracklst, pred ) {}
    };
    typedef FilteredTrackConstIterator<> TrackConstIterator;

    QString error;

    ITunesDB( IPodSysInfo& ipodSysInfo );

    virtual ~ITunesDB();

    /**
     * @brief Opens the itunesdb file at the given ipod mountpoint
     * @return true if the itunesdb does exist and could be opened. For parse errors check the error member afterwards
     */
    bool open();

    /**
     * @brief Returns true if open() has already been successfully called.
     * @return true if open has already been successfully called, false otherwise
     */
    bool isOpen();

    /**
     * @brief Writes the (changed) database back to the ipod
     * @param filename the name of the file to be written. If no filename is given the database will be written to the file read whith open()
     * @return true if writing the database was successful
     */
    bool writeDatabase(const QString& filename = QString());

    /**
     * @brief Returns true if the itunesdb file has been changed thus needs to be reloaded or does not exist anymore
     * @return true if the itunesdb file has been changed
     */
    bool dbFileChanged() const;

    /**
     * @brief Returns the name of the filename given with the open() call
     */
    QString getFilename() const { return itunesdbfile.name(); }

    /**
     * @brief Creates a new empty playlist
     */
    ITunesDBPlaylist * createNewPlaylist( const QString& title );

    /**
     * @brief Creates a new empty track
     *
     * Only use this function if you know what you're doing.
     * @see createNewTrack() for the right way to create a new Track
     * @param trackid the id of the new track in the database.
     */
    ITunesDBTrack * createEmptyTrack( Q_UINT32 trackid );

    /**
     * @brief Returns a new itunesdb::Track instance for adding a new track to the database.
     *
     * The new instance gets initialized with an unused trackid and trackpath.
     * Subsequent calls will get itunesdb::Track with the same trackid until the returned track is as added
     * via addTrack() thus a call to addTrack() has to be done afterwards.
     * @param fileExtension file extension of the file representing the track. If you don't know the file extension yet leave empty and set it with Track::setFileExtension() later
     */
    ITunesDBTrack * createNewTrack( const QString& fileExtension = QString::null );

    /**
     * @brief Adds a new track to the database and returns true if successfull.
     *
     * This only adds the control information to the database, the track itself (the file) has to by copied by the calling application.
     * If successfull the given pointer's owner is the database, if not the pointer needs to be deleted.
     * @param track the track to add to the database
     * @return true if the track was successfully added
     */
    bool addTrack( ITunesDBTrack * track );

    /**
     * @brief Returns the Track corresponding to the given ID
     * @param id ID of the track
     * @return the Track corresponding to the given ID
     */
    ITunesDBTrack * getTrackByID( const Q_UINT32 id ) const;

    /**
     * @brief Returns the first track found with the given dbid
     * @param dbid the 64bit DBID being unique for all objects in the database
     * @return the Track corresponding to the given DBID
     */
    ITunesDBTrack * findTrackByDBID( const Q_UINT64 dbid ) const;

    /**
     * @brief Returns the MHIT position of the track with the given ID
     * @param id ID of the track
     * @return the position of the track in the MHIT section or -1 if there's no such track.
     */
    int getMHITTrackPosFor( Q_UINT32 id ) const;

    /**
     * @brief Returns the Track at the specified position in the itunesdb
     * @param mhitTrackPos the position of the desired track
     * @return the track at the specified position or NULL if the given position is out of range
     */
    ITunesDBTrack * getTrackAt( uint mhitTrackPos) const;

    /**
     * @brief Returns the first Track where the given predicate returned true
     *
     * The given trackpredicate needs to contain a method that gets a track, and returns a bool
     * if it's the desired track
     * @return the first Track where the given predicate returned true
     * @see TrackPredicate for details about TrackPredicates
     */
    ITunesDBTrack * findFirstTrackBy( const itunesdb::TrackPredicate& predicate ) const;

    /**
     * @brief Returns all the Tracks where the given @c itunesdb::TrackPredicate returned true
     *
     * More formally the method fills in the Tracks where the given @c itunesdb::TrackPredicate
     * returned true into the given TrackPtrList and returns a pointer to the list if at least
     * one track got added.
     * @see getTracksBy() function below for a better way to get access to a sequence of tracks.
     * The returned iterator will deliver ITunesDBTrack instances with some extra functionalities.
     * @see itunesdb::TrackPredicates for a list of predefined predicates
     */
    itunesdb::TrackPtrList * getTracksBy( const itunesdb::TrackPredicate& predicate, itunesdb::TrackPtrList& buffer ) const;

    /**
     * @brief Returns all the Tracks where the given @c itunesdb::TrackPredicate returns true
     *
     * The next() method of the returned Iterator delivers pointers to ITunesDBTrack instances
     * @see itunesdb::TrackPredicates for a list of predefined predicates
     * @return an iterator over the Tracks where the given @c itunesdb::TrackPredicate returns true
     */
    template <typename TrackPredicate_T>
    inline FilteredTrackConstIterator<TrackPredicate_T> getTracksBy( const TrackPredicate_T& predicate ) const {
        return FilteredTrackConstIterator<TrackPredicate_T>( m_TrackList, predicate );
    }

    /**
     * @brief Returns the Track found by the given information or NULL if no such Track could be found
     * @param artistname the name of the artist
     * @param albumname the name of the album
     * @param title the title of the track
     * @param trackNum the position of the track in its album
     * @return the Track found or NULL if no track matched the given criteria.
     */
    ITunesDBTrack * findTrack(const QString& artistname, const QString& albumname, const QString& title, Q_UINT32 trackNum = 0 ) const;

    /**
     * @brief Fills the given QStringList with the names of all artists in the database
     * @param buffer a QStringList to fill in the artist names
     * @return a pointer to the StringList
     */
    QStringList * getArtists( QStringList& buffer ) const;

    /**
     * @brief Fills the given QStringList with the names of all albums in the database
     */
    QStringList * getAllAlbums( QStringList& buffer ) const;

    /**
     * @brief Fills the given QStringList with the names of the albums by the given artist
     * @param artistname the name of the artist to look for
     * @param buffer the list to put the names of the albums found
     * @return a pointer to the given QStringList or NULL if no such artist exists
     */
    QStringList * getAlbumsByArtist( const QString& artistname, QStringList& buffer ) const;

    /**
     * @brief Returns true if the given artist exists.
     * @param artistname the artists name to check for.
     * @return true if the artist exists, false otherwise.
     */
    bool containsArtist( const QString& artistname ) const;

    /**
     * @brief Returns true if the given album exists.
     * @param artistname the artists name to check for
     * @param albumname the name of the album
     * @return true if such an album exists for the given artist, false otherwise.
     */
    bool containsAlbum( const QString& artistname, const QString& albumname ) const;

    /**
     * @brief Fills the given TrackPtrList with all itunesdb::Track instances in the database.
     *
     * Fills the given TrackPtrList with all itunesdb::Track instances in the database and returns
     * a pointer to the given list.
     * @see getAllTracks() function below for a better way to get access to a sequence of tracks.
     * The returned iterator will deliver ITunesDBTrack instances with some extra functionalities.
     * @param buffer the list to put the tracks into.
     * @return the pointer to the given filled list
     */
    itunesdb::TrackPtrList * getAllTracks( itunesdb::TrackPtrList& buffer ) const;

    /**
     * @brief Returns an Iterator over all the tracks in the database
     *
     * @return an Iterator over all the tracks in the database
     */
    inline TrackConstIterator getAllTracks() const {
        return ITunesDB::TrackConstIterator( m_TrackList );
    }

    /**
     * @brief Returns an album by the given artistname and title.
     *
     * This method fills in all tracks matching the given artist/album information to the given TrackPtrList
     * @see getAlbum() function below for a better way to get access to a sequence of tracks.
     * The returned iterator will deliver ITunesDBTrack instances with some extra functionalities.
     * @param artistname name of the artist of the album in question
     * @param albumname name of the album
     * @param buffer the list the tracks of the album will be put to
     * @return a pointer to the given buffer or NULL if no such album could by found
     */
    itunesdb::TrackPtrList *
        getAlbum( const QString& artistname, const QString& albumname, itunesdb::TrackPtrList& buffer ) const;

    typedef FilteredTrackConstIterator<itunesdb::TrackPredicates::ByAlbum> AlbumIterator;
    /**
     * @brief Returns all the tracks matching the given artist and album
     * @param artistname name of the artist of the album in question
     * @param albumname name of the album
     * @return an iterator over the tracks matching the given artist and album
     */
    inline AlbumIterator getAlbum( const QString& artistname, const QString& albumname ) const {
        return getTracksBy( itunesdb::TrackPredicates::ByAlbum( artistname, albumname ) );
    }

    /**
     * @brief Returns all the Tracks by the given artist.
     *
     * This method fills in all tracks matching the artist to the given TrackPtrList
     * @see getTracksByArtist() function below for a better way to get access to a sequence of tracks.
     * The returned iterator will deliver ITunesDBTrack instances with some extra functionalities.
     * @param artistname the name of the artist to look for.
     * @param buffer the list to put the found tracks into.
     * @return the pointer to the given buffer or NULL if no such artist exists.
     */
    itunesdb::TrackPtrList *
        getTracksByArtist( const QString& artistname, itunesdb::TrackPtrList& buffer ) const;


    typedef FilteredTrackConstIterator<itunesdb::TrackPredicates::ByArtist> ArtistIterator;
    /**
     * @brief Returns all the tracks matching the given artistname
     * @param artistname the name of the artist to look for.
     * @return an iterator over the tracks matching the given artist name
     */
    inline ArtistIterator getTracksByArtist( const QString& artistname ) const {
        return getTracksBy( itunesdb::TrackPredicates::ByArtist( artistname ) );
    }

    /**
     * @brief Returns all tracks with the given string in its metadata
     *
     * This method fills in all tracks matching the given String in its metadata (artist, album, title or composer)
     * to the TrackPtrList.
     * @see getTracksContaining() function below for a better way to get access to a sequence of tracks.
     * The returned iterator will deliver ITunesDBTrack instances with some extra functionalities.
     * @param someString the string to look for
     * @param buffer the list to put the found tracks into
     * @param caseSensitive set to false if uppercase/lowercase should be ignored
     * @return the pointer to the given list or NULL if no such tracks exist.
     */
    itunesdb::TrackPtrList *
        getTracksContaining( const QString& someString, itunesdb::TrackPtrList& buffer, bool caseSensitive = true ) const;

    typedef FilteredTrackConstIterator<itunesdb::TrackPredicates::Contains> ContainsIterator;
    /**
     * @brief Returns all tracks with the given string in its metadata (artist, album, title or composer)
     * @param someString the string to look for
     * @param caseSensitive set to false if uppercase/lowercase should be ignored
     * @return an iterator over the tracks matching the given String name
     */
    inline ContainsIterator getTracksContaining( const QString& someString, bool caseSensitive = true ) const {
        return getTracksBy( itunesdb::TrackPredicates::Contains( someString, caseSensitive ) );
    }

    /**
     * @brief Returns a list with all playlist names
     * @param buffer the list to put the playlisttitles into
     * @return the pointer to the given buffer
     */
    QStringList * getPlaylistTitles( QStringList& buffer ) const;

    /**
     * @brief Returns a playlist by the given playlist title.
     * @param playlisttitle title of the playlist in question
     * @return a pointer to the playlist found or NULL if no such playlist exists
     */
    ITunesDBPlaylist * getPlaylistByTitle(const QString& playlisttitle) const;

    /**
     * @brief Returns the playlist by the given playlistID.
     * @param playlistID the DB ID of the desired playlist
     * @return the playlist found or NULL if no such playlist could be found.
     */
    ITunesDBPlaylist * getPlaylistByID( Q_UINT64 playlistID ) const;

    /**
     * @brief Removes all the tracks of the given artist (an thus the artist itself)
     * from the database.
     * @param artistname the name of the artist
     * @return true if tracks were found and removed, false otherwise.
     */
    bool removeArtist(const QString& artistname);

    /**
     * @brief Renames an album with the given title and artist
     *
     * Renames an album with the given title and artist to newtitle, newartist by
     * changing the artist and album properties of all the tracks found by the
     * given artist/album combination to newartist/newalbum.
     * @param artist the artist to look for
     * @param title the album to look for
     * @param newartist the new artist name to set. Give an empty or QString::null string if the artist property should remain unchanged.
     * @param newtitle the new album title to set. Give an empty or QString::null string if the album property should remain unchanged.
     * @return true if tracks were found and changed, false otherwise
     */
    bool renameAlbum(const QString& artist, const QString& title, const QString& newartist, const QString& newtitle);

    /**
     * @brief Adds a new playlist to the database.
     * @param playlist The playlist to be added to the database.
     * @return false if a playlist with the same title already exists, true otherwise
     */
    bool addPlaylist( ITunesDBPlaylist * playlist );

    /**
     * @brief Removes the playlist with the given title
     * @param title the title of the playlist to be removed
     * @param delete_instance if set to true the element gets deleted during removal
     * @return true if successful, otherwise false
     */
    bool removePlaylist( const QString& title, bool delete_instance);

    /**
     * @brief Updates all Smart Playlists in this itunesdb instance.
     */
    void updateSmartPlaylists();

    /**
     * @brief Removes the Track with the given trackid from the database.
     *
     * This also removes all playlist references to this track
     * @param trackid ID of the track to be deleted
     * @param delete_instance if set to true (default) the Track gets deleted during removal
     * @return the trackid of the Track or 0 if no such Track exists
     */
    Q_UINT32 removeTrack(Q_UINT32 trackid, bool delete_instance = true);

    /**
     * @brief Wipes all data from the container.
     */
    void clear();

    /**
     * @brief Returns true if something in the container's control has been changed.
     *
     * A change to a Playlist or a Track for example will not set the change flag yet.
     * @return true if something in the container's control has been changed.
     */
    bool isDirty() const;

    /**
     * @brief Returns the number of tracks
     * @return the number of tracks
     */
    Q_UINT32 getNumTracks() const;

    /**
     * @brief Returns the number of playlists
     * @return the number of playlists
     */
    Q_UINT32 getNumPlaylists() const;

    /**
     * @brief Lock control functions to prevent concurrent access to the Database from different instances
     */
    void lock(bool write_lock);
    void unlock();
    bool isLocked() const;

    /**
     * @brief Returns the timestamp the database file was modified.
     */
    uint lastModified() const { return timestamp.toTime_t(); }

    /**
     * @brief sets the title of the database
     *
     * The title will be stored as the title of the main track list
     * @param title the new title for the database
     */
    void setTitle( const QString& title );

    /**
     * @brief Returns the title of the database
     * @return the title of the main database
     */
    const QString& getTitle() const { return m_Title; }

    /**
     * @brief Returns true if the database contains podcasts
     * @return true if the database contains podcasts
     */
    bool hasPodcasts() const { return hasPodcastsFlag; }

    /**
     * @brief Returns the next free trackID for a new track
     * @return the next free trackID for a new track
     */
    Q_UINT32 getNextFreeTrackID() const;

    /**
     * @brief Returns the next available DBID.
     * @return the next available DBID.
     */
    Q_UINT64 getNextFreeDBID() const;

    /**
     * @brief Returns the fully qualified path for the given ipod path information
     *
     * Translates the pathinfo stored on the ipod to a real path info and adds the
     * ipodBase path so we can access the real file. The caller needs to check if
     * the file actually exists.
     * @param pathinfo the path property in database items
     * @return the path pointing to the file.
     */
    QString getFileForPathInfo( QString pathinfo ) const;

    /**
     * @brief Returns the ipod's base directory.
     * @return the ipod's base directory.
     */
    const QString& getIPodBase() const {
        return m_ipodSysInfo.getDeviceMountpoint();
    }

protected:

    TrackIDMap m_TrackIdMap;
    TrackList m_TrackList;
    QString m_Title;

    /**
     * The container that holds all the playlist for this iTunesDB instance.
     */
    class PlaylistContainer : public std::vector<ITunesDBPlaylist *> {
    public:
        PlaylistContainer() : std::vector<ITunesDBPlaylist *>() {}
        virtual ~PlaylistContainer() { clear(); }
    } * m_playlists;

    bool m_dirty;
    void setDirty( bool dirtyFlag = true );

    Q_UINT32 maxtrackid;
    Q_UINT64 mMaxDBID;

    ITunesDBTrack * insertTrackToDataBase( ITunesDBTrack * track );

private:

    void removeFromAllPlaylists( Q_UINT32 trackid );
    IPodSysInfo& m_ipodSysInfo;
    QFile itunesdbfile;
    QFile itunessdfile;
    QDateTime timestamp;

    bool hasPodcastsFlag;
    bool locked;
};


#endif
