 /*****************************************************************************
 *   Copyright (C) 2006-2007 by Michael Schulze                               *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#ifndef _qtpod_itunesdb_utils_H_
#define _qtpod_itunesdb_utils_H_

// here you can enable usage of the boost shared_ptr impl if using an older libstdc++
// uncomment the following line to use the boost shared_ptr
// #define QTPOD_USE_BOOST_SHAREDPTR 1

#ifdef QTPOD_USE_BOOST_SHAREDPTR

#include <boost/shared_ptr.hpp>
#define QTPOD_SHARED_PTR_IMPL_DEF boost::shared_ptr

#else

#include <tr1/memory>
#define QTPOD_SHARED_PTR_IMPL_DEF std::tr1::shared_ptr

#endif

#include <vector>

#define MAC_EPOCH_DELTA 2082844800

#include <algorithm>

namespace itunesdb {

namespace utils {

static const QString NullQString = QString();

/**
 * some stuff to find/filter etc through containers
 */

/**
 * Returns the Iterator where predicate(*iter) == TRUE
 */
template <class Iter, class TUnaryPredicate>
Iter findFirst( Iter pos, Iter end, const TUnaryPredicate& predicate ) {
    for (; pos != end; ++pos) {
        if (predicate(*pos)) {
            return pos;
        }
    }
    return end;
}

/**
 * Returns the Iterator to the element where the given function returned the biggest result
 */
template <class Iter, class TUnaryFunction>
Iter max( Iter pos, Iter end, const TUnaryFunction& fn ) {
    Iter currentMax = pos;
    for ( ; pos != end; ++pos ) {
        if ( fn( *pos ) > fn ( *currentMax ) ) {
            currentMax = pos;
        }
    }
    return currentMax;
}


/**
 * A marker predicate for unfiltered Iterators
 */
struct TRUEPredicate {};

/*
 * Internal functions for filtered Iterators
 */
template < typename IterType, typename Iter, typename TUnaryPredicate, typename TDereferenceFun >
struct RangeIteratorFunctions {
    TUnaryPredicate pred;
    TDereferenceFun deref;

    RangeIteratorFunctions( const TUnaryPredicate& p, const TDereferenceFun& d ) :
        pred( p ), deref( d ) {}

    Iter successor( const Iter& current, const Iter& end ) const {
        Iter result = current;
        while ( ++result != end && !pred( deref( result ) ) );
        return result;
    }

    bool isValid( const Iter& i ) const { return pred( deref( i ) ); }

    unsigned int remaining( const Iter& next, const Iter& end ) {
        Iter i = next;
        unsigned int result = 0;
        while ( i != end ) {
            result++;
            i = successor( i, end );
        }
        return result;
    }
};

/*
 * Internal functions for unfiltered Iterators
 */
template < typename IterType, typename Iter, typename TDereferenceFun >
struct RangeIteratorFunctions<IterType, Iter, TRUEPredicate, TDereferenceFun> {
    RangeIteratorFunctions( const TRUEPredicate&, const TDereferenceFun& ) {}

    Iter successor( const Iter& current, const Iter& ) const {
        Iter result = current;
        return ++result;
    }

    bool isValid( const Iter& ) const { return true; }

    unsigned int remaining( const Iter& next, const Iter& end ) {
        return end - next;
    }
};

/**
 * Default dereferencing functor just dereferencing iterators
 */
template <typename PType, typename Iter> struct DefaultDeref {

    /**
     * Returns the dereferenced Iter
     * @param iter the Iter to dereference
     * @return the dereferenced Iter
     */
    const PType operator()( const Iter& iter ) const {
        return *iter;
    }
};

/**
 * @brief An Iterator over a range of elements (in most cases backed by some container).
 *
 * The RangeIterator can be used to iterate over a collection of elements using its hasNext() and next() methods.
 * For example to iterate over all the Tracks of a @c itunesdb::TrackPtrList :
 * <pre><code>
    itunesdb::TrackPtrList::Iterator trackIter = tlist.iterator();
    while( trackIter.hasNext() ) {
        itunesdb::Track * track = trackIter.next();
        ... some code ...
    }
 * </code></pre>
 * The TUnaryPredicate decides if an element is part of the range and therefore should be taken into account by hasNext()
 * and next() or not.
 * The Dereference Function dereferences an iterator and returns IterType. That's helpful if you need to create
 * an iterator over the keys of a map without being too obvious about your implementation details.
 * @see DereferencingRangeIterator for an example dereferencing the keys of a map.
 * @author Michael Schulze
 */
template <typename IterType, typename Iter, typename TUnaryPredicate = TRUEPredicate, typename TDereferenceFun = DefaultDeref< IterType, Iter > >
class RangeIterator {
    Iter m_current; // points to the element returned by next
    Iter m_next;    // points to the element to be returned by next()
    Iter m_end; // one beyond the last to return

protected:

    /**
     * Dereferencer to dereference an iterator to IterType
     */
    TDereferenceFun m_dereferenceFun;

    /**
     * Internal implementors of the underlying functions
     */
    RangeIteratorFunctions<IterType, Iter, TUnaryPredicate, TDereferenceFun> m_helper;

    /**
     * @brief Sets the range to the given iterators
     */
    void setRange( Iter pos, Iter end ) {
        m_end = end;
        m_current = pos;
        if ( m_current != m_end && !m_helper.isValid( m_current ) ) {
            m_current = m_helper.successor( m_current, m_end );
        }
        m_next = m_current;
    }

    /**
     * @brief  Returns the iterator pointing to the element returned by the last next() call
     * @return the iterator pointing to the element returned by the last next() call
     */
    Iter currentPos() {
        return m_current;
    }

    /**
     * @brief Returns true if there are no elements left to be iterated over
     * @return true if there are no elements left to be iterated over
     */
    bool empty() const {
        return m_current == m_end;
    }

public:

    /**
     * @brief Constructs a new Iterator from start to end with the given predicate.
     */
    RangeIterator(Iter start, Iter end, const TUnaryPredicate& predicate = TUnaryPredicate(), TDereferenceFun deref = TDereferenceFun() )
        : m_dereferenceFun( deref ), m_helper( predicate, deref )
    {
        setRange( start, end );
    }

    /**
     * @brief Returns true if there are elements left so calling next() would return the next element rather than causing a segfault
     * @return true if there are elements left
     */
    bool hasNext() const {
        return m_next != m_end;
    }

    /**
     * @brief Returns the next element of the range of elements we iterate over.
     *
     * This method positions the Iterator at the next element and returns it. The first call to
     * this method will return the first element of the range.
     * @n If the iterator is filtered only those elements where the given predicate returned true
     * are returned.
     * @attention Always check if hasNext() returns true before calling next. Calling next() on an iterator already at
     * the end of the range will cause a segfault.
     * @pre hasNext() returns true
     * @post a successive call to current() returns the element returned by this method
     */
    IterType next() {
        m_current = m_next;
        m_next = m_helper.successor( m_current, m_end );
        return current();
    }

    /**
     * @brief Returns the element returned by the last next() call
     * @return the element returned by the last next() call
     */
    IterType current() const {
        return m_dereferenceFun( m_current );
    }

    /**
     * @brief Returns the number of elements remaining in this iterator.
     *
     * ... meaning the number of times the next() method can be called before the hasNext() method willreturn false.
     * For filtered iterators this may be a lengthy operation since the iterator needs to apply its filter over all
     * elements to determine how many elements are left.
     * @return the number of elements left
     */
    unsigned int remaining() const {
        return m_helper.remaining( m_next, m_end );
    }

    /**
     * @brief Returns the last element this iterator would return.
     *
     * This method is dangerous and makes no sense. Do not call this for empty iterators so at least check with hasNext() before.
     * @return the last element this iterator would return
     * @pre it must have been ensured that the iterator is nonempty
     */
    IterType last() const {
        Iter result = m_end;
        while( result != m_current && !m_helper.isValid( --result ) );
        return m_dereferenceFun( result );
    }

};

/**
 * @brief An unfiltered RangeIterator with a custom dereferencer.
 *
 * Use this method if you don't want to return just *iter in your next() method. This is useful if you want to iterate
 * over the keys of a map where you'd return iter->first in that case.
 * The key dereferencer for a std::map with a key type of QString would look something like this:
 * <pre><code>
    struct KeyDereferencer {
        const QString& operator() ( std::map<QString,whatever>::const_iterator& iter ) const {
            return (*iter).first;
        }
    };
 * </code></pre>
 * A KeyIterator would then look something loke this:
 * <pre><code>
    typedef itunesdb::utils::DereferencingRangeIterator<QString,std::map<QString,whatever>::const_iterator, KeyDereferencer> KeyIterator;
 * </code></pre>
 * and its instantiation
 * <pre><code>
    KeyIterator keyIter( map.begin(), map.end() );
 * </code></pre>
 * To print out the keys of the map we use the keyIter above like this
 * <pre><code>
    while( keyIter.hasNext() ) {
        printf( "%s\n", keyIter.next().ascii() );
    }
 * </code></pre>
 */
template <typename IterType, typename Iter, typename TDereferenceFun = DefaultDeref<IterType, Iter> >
class DereferencingRangeIterator : public RangeIterator<IterType, Iter, TRUEPredicate, TDereferenceFun> {
public:

    /**
     * Creates a DereferencingRangeIterator over the given range.
     */
    DereferencingRangeIterator( Iter start, Iter end )
        : RangeIterator< IterType, Iter, TRUEPredicate, TDereferenceFun >( start, end, TRUEPredicate() )
    {
    }
};


template <class Comparable_T>
class Comparator {
public:
    virtual ~Comparator() {}
    /**
     * Compares the two given comparables and returns a positive, 0 or a negative value
     * if the first comparable is greater than, equal to, or less than the second.
     * @param c1 the first comparable to compare
     * @param c2 the second comparable to compare
     * @return a positive, 0 or a negative value if the first comparable is greater than, equal to, or less than the second.
     */
    virtual int compare( const Comparable_T& c1, const Comparable_T& c2 ) const = 0;
};


/**
 * A Class implementing a sortable Vector storing pointers to the given Type
 */
template <class ElemType>
class SortablePtrVector : protected std::vector<ElemType*> {

    typedef typename std::vector<ElemType*>::iterator PrivateIterator;
    typedef typename std::vector<ElemType*>::const_iterator PrivateConstIterator;
    typedef std::vector<ElemType*> BaseContainerType;

    unsigned long m_version;

public:

    /**
     * @brief Comparator Type definition
     */
    typedef Comparator<ElemType> SortableVectorElemComparator;

    /**
     * @brief Shared Pointer type managing the pointer to the comparator
     */
    typedef QTPOD_SHARED_PTR_IMPL_DEF< SortableVectorElemComparator > ComparatorPtr;

     /**
     * An Iterator over all elements of this container being aware of changes to the container.
     * The Iterator gets invalid if the container got changed outside the control of this iterator.
     */
    template < typename Container_T, typename Iter_T, typename TUnaryPredicate = TRUEPredicate >
    class ContainerVersionAwareIterator : public RangeIterator<ElemType*, Iter_T, TUnaryPredicate > {
    protected:

        /**
         * @brief The container we iterate over.
         *
         * We need this to check our version and validity against
         */
        Container_T& m_container;

        /**
         * @brief The version of the container we iterate over.
         *
         * If this differs from m_container.m_version the iterator gets invalid.
         */
        unsigned long m_containerversion;

        /**
         * @brief The base type of this iterator
         */
        typedef RangeIterator< ElemType*, Iter_T, TUnaryPredicate > BaseRangeIterator;

        /**
         * @brief Creates a new ContainerVersionAwareIterator
         *
         * The Iterator iterates over all elements in container matching the given predicate.
         * The version of the iterator will be checked against the version of the given container
         * to check the validity of this Iterator instance.
         * @param container the container to iterate over and to check the version against
         * @param pred the predicate to match the elements against
         */
       ContainerVersionAwareIterator( Container_T& container, const TUnaryPredicate& pred = TUnaryPredicate() )
            : BaseRangeIterator( container.begin(), container.end(), pred ),
              m_container( container ),
              m_containerversion( container.m_version ) {}

    public:

        /**
         * @brief Returns false if the underlying container got changed outside the control of this iterator, true otherwise.
         * @return false if the underlying container got changed outside the control of this iterator.
         */
        bool isValid() const {
            return m_container.m_version == m_containerversion;
        }

        /**
         * @brief Returns true if there are elements left so the next() may be called to get the next element
         * @n This method will always return false if the underlying container got changed and thus this iterator got invalidated.
         * @return true if there are elements left so the next() may be called to get the next element
         * @overload RangeIterator::hasNext()
         */
        bool hasNext() const {
            return isValid() && BaseRangeIterator::hasNext();
        }

    };

   /**
     * A non const @c ContainerVersionAwareIterator providing functions to change the underlying container.
     */
    template < typename TUnaryPredicate = TRUEPredicate >
    class FilteredIterator : public ContainerVersionAwareIterator<SortablePtrVector, PrivateIterator, TUnaryPredicate > {
        friend class SortablePtrVector;

        typedef ContainerVersionAwareIterator<SortablePtrVector, PrivateIterator, TUnaryPredicate > IteratorBase;

        unsigned long incrementContainerVersion() {
            if ( IteratorBase::isValid() ) {
                return IteratorBase::m_containerversion = ++IteratorBase::m_container.m_version;
            }
            return 0;
        }

    public:

        /**
         * @brief Creates a new FilteredIterator over all elements in container matching the given predicate.
         * @param container the container to iterate over
         * @param predicate the predicate to match the elements against
         */
        FilteredIterator( SortablePtrVector& container, const TUnaryPredicate& predicate = TUnaryPredicate() )
            : IteratorBase( container, predicate ) {}

        /**
         * Removes the element returned by the last next() call from the container and positions the iterator at the next element.
         * @return true if the Iterator is valid and non empty, false otherwise
         */
        bool remove() {
            if ( IteratorBase::empty() || !IteratorBase::isValid() ) {
                return false;
            }
            incrementContainerVersion();
            ElemType * element = IteratorBase::current();
            PrivateIterator newpos = IteratorBase::m_container.erase( IteratorBase::currentPos() );
            IteratorBase::setRange( newpos, IteratorBase::m_container.end() );
            if ( IteratorBase::m_container.m_deleteElements ) {
                delete element;
            }
            return true;
        }

        /**
         * Removes the elements from the current position until the end that are part of this RangeIterator.
         * For filtered iterators only those elements matching the filter get removed.
         * @return true if at least one element got removed
         */
        bool removeRemaining() {
            if ( !IteratorBase::hasNext() ) {
                return false;
            }
            do {
                IteratorBase::next();
                remove();
            } while ( IteratorBase::hasNext() );
            return true;
        }
    };

    /**
     * An unfiltered Iterator over all elements of this container.
     */
    typedef FilteredIterator<> Iterator;

    /**
     * A filtered const Iterator over the elements of this container where the given predicate returned true.
     * @see @c RangeIterator for more details
     */
    template < typename TUnaryPredicate = TRUEPredicate >
    struct FilteredConstIterator : public ContainerVersionAwareIterator<const SortablePtrVector, PrivateConstIterator, TUnaryPredicate > {
        friend class SortablePtrVector;

        /**
         * @brief Creates a new FilteredConstIterator over all elements in container matching the given predicate.
         * @param container the container to iterator over
         * @param predicate the predicate to match the elements against
         */
        FilteredConstIterator( const SortablePtrVector& container, const TUnaryPredicate& predicate = TUnaryPredicate() )
            : ContainerVersionAwareIterator<const SortablePtrVector, PrivateConstIterator, TUnaryPredicate >( container, predicate ) {}
    };

    /**
     * A const Iterator over all elements of this container.
     * @see @c RangeIterator for more details
     */
    typedef FilteredConstIterator<> ConstIterator;

private:
    bool m_deleteElements;
    ComparatorPtr m_comparator;

public:
    /**
     * Creates a sortable vector storing pointers to its elements.
     * @param deleteElements if true the container will delete elements being removed from this container.
     */
    SortablePtrVector( bool deleteElements = false )
        : BaseContainerType(), m_version( 0 ), m_deleteElements( deleteElements ) {}

    /**
     * @brief Creates a sortable vector storing pointers to its elements and fills in the objects returned by the given iterator.
     * @param elements the elements to be filled in to the container
     * @param deleteElements if true the container will delete elements being removed from this container.
     */
    template < typename IterT >
    SortablePtrVector( IterT elements, bool deleteElements = false )
        : BaseContainerType(), m_version( 0 ), m_deleteElements( deleteElements )
    {
        addAll( elements );
    }

    virtual ~SortablePtrVector() {
        clear();
    }

    /**
     * @brief Sets the comparator for this container to by used by sort() and inSort()
     * @param comparator the comparator to be set. The pointer is owned by the container and will be deleted if a new comparator is set.
     */
    void setComparator( SortableVectorElemComparator * comparator ) {
        m_comparator = ComparatorPtr( comparator );
    }

    /**
     * @brief Sets the comparator for this container to by used by sort() and inSort()
     * @param comparator the comparator to be set. The comparators lifecycle is managed by the ComparatorPtr.
     */
    void setComparator( const ComparatorPtr& comparator ) {
        m_comparator = comparator;
    }

    /**
     * @brief Returns the currently used comparator for sorting.
     * @return the currently used comparator for sorting.
     */
    const ComparatorPtr& getComparator() const {
        return m_comparator;
    }

    /**
     * @brief Sorts the containers element using the comparator set by setComparator().
     *
     * If no comparator has been set the order is kept unchanged.
     */
    void sort() {
        if ( m_comparator ) {
            ++m_version;    // invalidate iterators
            std::sort( BaseContainerType::begin(), BaseContainerType::end(), SmallerBinaryPredicate( m_comparator ) );
        }   // else leave it as it is
    }

    /**
     * @brief Randomizes the order of the elements.
     */
    void randomize() {
        ++m_version;    // invalidate iterators
        random_shuffle( BaseContainerType::begin(), BaseContainerType::end() );
    }

    struct SmallerBinaryPredicate {
        ComparatorPtr m_comparator;
        SmallerBinaryPredicate( ComparatorPtr& ptr ) : m_comparator( ptr ) {}
        bool operator() ( const ElemType * i1, const ElemType * i2 ) {
            return m_comparator->compare( *i1, *i2 ) < 0;
        }
    };
    struct FindFirstNotLessThan {
        SmallerBinaryPredicate& m_predicate;
        ElemType * m_item;
        FindFirstNotLessThan( SmallerBinaryPredicate& pred, ElemType * item )
            : m_predicate( pred ), m_item( item ) {}
        bool operator() ( const ElemType * i ) {
            return m_predicate( m_item, i );
        }
    };
    /**
     * @brief Inserts the given element accordingly to the comparator set by setComparator().
     *
     * If no comparator has been set the element will be appended at the end.
     * @param item the element to be inserted
     * @return the position/index of the item in this vector.
     */
    unsigned int inSort( ElemType * item ) {
        unsigned int idx = 0;

        ++m_version;

        if ( !m_comparator ) {
            idx = BaseContainerType::size();
            BaseContainerType::push_back( item );
            return idx;
        }

        SmallerBinaryPredicate pred( m_comparator );
        FindFirstNotLessThan finder( pred, item );

        PrivateIterator pos = find_if( BaseContainerType::begin(), BaseContainerType::end(), finder );
        if ( pos == BaseContainerType::end() ) {
            idx = BaseContainerType::size();
            BaseContainerType::push_back( item );
        } else {
            idx = pos - BaseContainerType::begin();
            BaseContainerType::insert( pos, item );
        }

        return idx;
    }


    /**
     * @brief Adds all elements delivered by the given Iterator.
     */
    template <typename IterT>
        void addAll( IterT elemIter ) {
            while ( elemIter.hasNext() ) {
                append( elemIter.next() );
            }
        }

    /**
     * @brief Removes the element at the given position from the container.
     *
     * If the container was created with deleteElements set to true the element will be deleted.
     * @param pos the position of the element to be removed.
     * @return true if the given position was valid accordingly to the number of elements.
     */
    bool remove( unsigned int pos ) {
        if ( pos > BaseContainerType::size() -1 ) {
            return false;
        }

        ++m_version;

        PrivateIterator posIter = BaseContainerType::begin() + pos;
        if ( m_deleteElements ) {
            delete *posIter;
        }
        this->erase( posIter );

        return true;
    }

    /**
     * @brief Returns the number of elements,
     * @return the number of elements,
     */
    unsigned int count() const { return BaseContainerType::size(); }

    /**
     * @brief Returns true if the number of elements is zero meaning the container is empty, false otherwise.
     * @return true if the container is empty.
     */
    bool isEmpty() const { return BaseContainerType::empty(); }

    /**
     * @brief Returns a @c SortablePtrVector::Iterator over all elements of this Container
     * @return a @c SortablePtrVector::Iterator over all elements of this Container
     */
    Iterator iterator() { return Iterator ( *this ); }

    /**
     * @brief Returns a @c SortablePtrVector::ConstIterator over all elements of this Container
     * @return a @c SortablePtrVector::ConstIterator over all elements of this Container
     */
    ConstIterator iterator() const { return ConstIterator( *this ); }

    /**
     * @brief Returns a @c SortablePtrVector::ConstIterator over all elements of this Container
     * @return a @c SortablePtrVector::ConstIterator over all elements of this Container
     */
    ConstIterator const_iterator() const { return ConstIterator( *this ); }

    /**
     * @brief Returns a FilteredIterator over those elements matching the given predicate.
     * @param pred the predicate used for filtering
     * @return a filtered Iterator over the elements of this container.
     */
    template <typename TUnaryPredicate>
    FilteredIterator<TUnaryPredicate> filteredElements( const TUnaryPredicate& pred ) {
        return FilteredIterator<TUnaryPredicate>( *this, pred );
    }

    /**
     * @brief Returns a FilteredConstIterator over those elements matching the given predicate.
     * @param pred the predicate used for filtering
     * @return a filtered const Iterator over the elements of this container.
     */
    template <typename TUnaryPredicate>
    FilteredConstIterator<TUnaryPredicate> filteredElements( const TUnaryPredicate& pred ) const {
        return FilteredConstIterator<TUnaryPredicate>( *this, pred );
    }

    /**
     * @brief Appends the given element at the end of the container disregarding any comparator.
     * @param item the element to be appended.
     */
    void append( ElemType * item ) { BaseContainerType::push_back( item ); }

    /**
     * @brief An operator to access the element at the given position.
     * @param pos the position of the element to be returned
     * @return a pointer to the element at the given position
     */
    ElemType * operator[] ( unsigned int pos ) const {
        return BaseContainerType::operator[] ( pos );
    }

    /**
     * @brief An operator to access the element at the given position.
     * @param pos the position of the element to be returned
     * @return a reference to the pointer at the given position
     */
    ElemType *& operator[] ( unsigned int pos ) {
        return BaseContainerType::operator[] ( pos );
    }

    /**
     * @brief Returns a pointer to the last element.
     * @return a pointer to the last element.
     */
    ElemType * last() const {
        return BaseContainerType::back();
    }

    /**
     * @brief Removes all elements.
     */
    void clear() {
#if 0
        // FIXME remove iterator testcode before release
        iterator().removeRemaining();
        printf( "%d number of items left after clear\n" , count() );
#else
        if ( m_deleteElements ) {
            PrivateIterator iter = BaseContainerType::begin();
            for ( ; iter != BaseContainerType::end(); ++iter ) {
                delete *iter;
            }
        }
        BaseContainerType::clear();
#endif
    }
};


/**
 * @brief inherit from this to get a non copyable class
 */
class NonCopyAble {
private:
    NonCopyAble( const NonCopyAble& ) {}
protected:
    NonCopyAble() {}
public:
    virtual ~NonCopyAble() {}
};


/**
 * @brief Seeks the given number of bytes in the given stream
 */
template <class Stream_T>
inline void seekRelative( Stream_T& stream, unsigned int numbytes ) {
    if ( numbytes > 0 ) {
        char * buffer = new char[ numbytes ];
        stream.readRawBytes( buffer, numbytes );
        delete [] buffer;
    }
}


} // utils

} // itunesdb

#endif
